# -*- coding: utf-8 -*-

"""Setup and run MOPAC"""

import calendar
import datetime
import logging
import os
import os.path
from pathlib import Path
import pkg_resources
import pprint
import re
import string

import psutil

import molsystem
import seamm
import seamm_util
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __
import mopac_step

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("mopac")

# Add MOPAC's properties to the standard properties
path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "properties.csv"
molsystem.add_properties_from_file(csv_file)


class MOPAC(seamm.Node):
    def __init__(
        self, flowchart=None, namespace="org.molssi.seamm.mopac", extension=None
    ):
        """Initialize the node"""

        logger.debug("Creating MOPAC {}".format(self))

        # Create the subflowchart and proceed
        self.subflowchart = seamm.Flowchart(
            name="MOPAC",
            parent=self,
            namespace=namespace,
            directory=flowchart.root_directory,
        )
        self._data = {}
        self._lattice_opt = True

        super().__init__(
            flowchart=flowchart, title="MOPAC", extension=extension, logger=logger
        )

    @property
    def version(self):
        """The semantic version of this module."""
        return mopac_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return mopac_step.__git_revision__

    def create_parser(self):
        """Setup the command-line / config file parser"""
        parser_name = self.step_type
        parser = seamm_util.getParser()

        # Remember if the parser exists ... this type of step may have been
        # found before
        parser_exists = parser.exists(parser_name)

        # Create the standard options, e.g. log-level
        result = super().create_parser(name=parser_name)

        if parser_exists:
            return result

        # Options for Mopac
        parser.add_argument(
            parser_name,
            "--mopac-exe",
            default="mopac",
            help="the name of the MOPAC executable",
        )

        parser.add_argument(
            parser_name,
            "--mopac-path",
            default="",
            help="the path to the MOPAC executable",
        )

        parser.add_argument(
            parser_name,
            "--ncores",
            default="default",
            help="How many threads to use in MOPAC",
        )

        parser.add_argument(
            parser_name,
            "--mkl-num-threads",
            default="default",
            help="How many threads to use with MKL in MOPAC",
        )

        parser.add_argument(
            parser_name,
            "--max-atoms-to-print",
            default=25,
            help="Maximum number of atoms to print charges, etc.",
        )

        return result

    def set_id(self, node_id):
        """Set the id for node to a given tuple"""
        self._id = node_id

        # and set our subnodes
        self.subflowchart.set_ids(self._id)

        return self.next()

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """
        # Work through children. Get the first real node
        node = self.subflowchart.get_node("1").next()

        text = self.header + "\n\n"
        while node is not None:
            text += __(node.description_text(), indent=4 * " ").__str__()
            text += "\n"
            node = node.next()

        return text

    def run(self):
        """Run MOPAC"""
        system, configuration = self.get_system_configuration(None)
        n_atoms = configuration.n_atoms
        if n_atoms == 0:
            self.logger.error("MOPAC run(): there is no structure!")
            raise RuntimeError("MOPAC run(): there is no structure!")

        # Print our header to the main output
        printer.important(self.header)
        printer.important("")

        # Access the options and find the executable
        options = self.options
        seamm_options = self.global_options

        if options["mopac_path"] == "":
            mopac_exe = seamm_util.check_executable(options["mopac_exe"])
        else:
            mopac_exe = (
                Path(options["mopac_path"]).expanduser().resolve()
                / options["mopac_exe"]
            )
        mopac_path = Path(mopac_exe).parent.expanduser().resolve()

        # How many processors does this node have?
        n_cores = psutil.cpu_count(logical=False)
        self.logger.info("The number of cores is {}".format(n_cores))

        if seamm_options["ncores"] != "available":
            n_cores = min(n_cores, int(seamm_options["ncores"]))
        # Currently, on the Mac, it is not clear that any parallelism helps
        # much.

        # if options["ncores"] == "default":
        #     # Wild guess!
        #     # mopac_mkl_num_threads = int(pow(n_atoms / 16, 0.3333))
        #     mkl_num_threads = 1
        # else:
        #     if options["mkl_num_threads"] == "default":
        #         mkl_num_threads = n_cores
        #     else:
        #         mkl_num_threads = int(options["mkl_num_threads"])
        # if mkl_num_threads > n_cores:
        #     mkl_num_threads = n_cores
        # elif mkl_num_threads < 1:
        #     mkl_num_threads = 1
        # self.logger.info(f"MKL will use {mkl_num_threads} threads.")

        n_hydrogens = configuration.atoms.get_n_atoms("atno", "==", 1)
        n_basis = (n_atoms - n_hydrogens) * 4 + n_hydrogens
        if options["ncores"] == "default":
            # Since it is the matrix diagonalization, work out rough
            # size of matrix

            # Wild guess!
            mopac_num_threads = int(pow(n_basis / 1000, 3))
            if mopac_num_threads > n_cores:
                mopac_num_threads = n_cores
        else:
            mopac_num_threads = int(options["ncores"])
        if mopac_num_threads > n_cores:
            mopac_num_threads = n_cores
        if mopac_num_threads < 1:
            mopac_num_threads = 1
        self.logger.info(
            f"MOPAC will use {mopac_num_threads} threads for about {n_basis} basis "
            "functions."
        )

        env = {
            "LD_LIBRARY_PATH": str(mopac_path),
            "OMP_NUM_THREADS": str(mopac_num_threads),
        }
        # "MKL_NUM_THREADS": str(mkl_num_threads),

        # extra_keywords = ["AUX(MOS=10,XP,XS)", "NOXYZ"]
        extra_keywords = ["AUX(MOS=10,XP,XS)"]

        # Always add the charge since that will cause MOZYME, if used, to check.
        extra_keywords.append(f"CHARGE={configuration.charge}")
        # And the spin multiplicity
        multiplicity = configuration.spin_multiplicity
        if multiplicity <= 10:
            extra_keywords.append(
                (
                    "SINGLET",
                    "DOUBLET",
                    "TRIPLET",
                    "QUARTET",
                    "QUINTET",
                    "SEXTET",
                    "SEPTET",
                    "OCTET",
                    "NONET",
                )[multiplicity - 1]
            )
        else:
            extra_keywords.append(f"MS={(multiplicity - 1) / 2}")

        n_active_electrons = configuration.n_active_electrons
        n_active_orbitals = configuration.n_active_orbitals

        if n_active_orbitals > 0:
            extra_keywords.append(f"OPEN({n_active_electrons},{n_active_orbitals})")
            state = configuration.state
            if state != "1":
                extra_keywords.append(f"ROOT={state}")
        else:
            if multiplicity > 1:
                extra_keywords.append("UHF")

        # All Lanthanides (except La and Lu) must use the SPARKLES keyword.
        # La and Lu use the SPARKLES keyword optionally, depending
        # if you're looking for good structure (do use SPARKLES) or
        # energy (do not use SPARKLES)
        La = [
            "Ce",
            "Pr",
            "Nd",
            "Pm",
            "Sm",
            "Eu",
            "Gd",
            "Tb",
            "Dy",
            "Ho",
            "Er",
            "Tm",
            "Yb",
        ]

        La_list = set(La) & set(configuration.atoms.symbols)

        if len(La_list) > 0:
            extra_keywords.append("SPARKLES")

        # if mopac_num_threads > 1:
        #     extra_keywords.append("THREADS={}".format(mopac_num_threads))

        # Work through the subflowchart to find out what to do.
        self.subflowchart.root_directory = self.flowchart.root_directory

        next_node = super().run(printer)

        # Get the first real node
        node = self.subflowchart.get_node("1").next()

        text = ""
        n_calculations = []
        while node:
            node.parent = self
            inputs = node.get_input()
            n_calculations.append(len(inputs))
            for keywords, structure, comment in inputs:
                lines = []
                lines.append(" ".join(keywords + extra_keywords))
                lines.append(system.name)
                if comment is None:
                    lines.append(configuration.name)
                else:
                    lines.append(comment)

                text += "\n".join(lines)
                text += "\n"
                if structure is None:
                    if "OLDGEO" not in keywords:
                        text += self.mopac_structure()
                        text += "\n"
                else:
                    text += structure
                    text += "\n"
            node = node.next()

        files = {"mopac.dat": text}
        self.logger.debug("mopac.dat:\n" + files["mopac.dat"])
        os.makedirs(self.directory, exist_ok=True)
        for filename in files:
            with open(os.path.join(self.directory, filename), mode="w") as fd:
                fd.write(files[filename])
        local = seamm.ExecLocal()
        return_files = ["mopac.arc", "mopac.out", "mopac.aux"]
        result = local.run(
            cmd=[str(mopac_exe), "mopac.dat"],
            files=files,
            return_files=return_files,
            env=env,
        )

        if not result:
            self.logger.error("There was an error running MOPAC")
            return None

        self.logger.debug("\n" + pprint.pformat(result))

        self.logger.debug(
            "\n\nOutput from MOPAC\n\n" + result["mopac.out"]["data"] + "\n\n"
        )

        for filename in result["files"]:
            with open(os.path.join(self.directory, filename), mode="w") as fd:
                if result[filename]["data"] is not None:
                    fd.write(result[filename]["data"])
                else:
                    fd.write(result[filename]["exception"])

        # Analyze the results
        self.analyze(n_calculations=n_calculations)

        # Close the reference handler, which should force it to close the
        # connection.
        self.references = None

        return next_node

    def analyze(self, indent="", lines=[], n_calculations=None):
        """Read the results from MOPAC calculations and analyze them,
        putting key results into variables for subsequent use by
        other stages
        """

        # Split the aux files into sections for each step
        filename = "mopac.aux"
        with open(os.path.join(self.directory, filename), mode="r") as fd:
            lines_aux = fd.read().splitlines()

        # Find the sections in the file corresponding to sub-tasks
        # MOPAC keeps cumulative times, so fix them
        t_total = 0.0
        aux_data = []
        start = 0
        lineno = 0
        section = 0
        for line in lines_aux:
            if "END OF MOPAC FILE" in line:
                self.logger.debug("\nAUX file section {}".format(section))
                self.logger.debug("------------------")

                tmp_data = self.parse_aux(lines_aux[start:lineno])
                if "CPU_TIME" in tmp_data:
                    tmp = tmp_data["CPU_TIME"]
                    tmp_data["CPU_TIME"] = tmp - t_total
                    t_total = tmp
                aux_data.append(tmp_data)

                self.logger.debug(pprint.pformat(tmp_data, width=170, compact=True))
            lineno += 1
            if "START OF MOPAC FILE" in line:
                section += 1
                start = lineno

        # Split the output file into sections for each step
        filename = "mopac.out"
        with open(os.path.join(self.directory, filename), mode="r") as fd:
            lines = fd.read().splitlines()

        # Find the sections in the file corresponding to sub-tasks
        out = []
        start = 0
        lineno = 0
        for line in lines:
            if "** Cite this program as:" in line or "Digital Object Ident" in line:
                if lineno == 5:
                    continue
                out.append(lines[start : lineno - 5])
                start = lineno - 6
            lineno += 1
        out.append(lines[start:])

        for data in aux_data:
            # Add main citation for MOPAC
            if "MOPAC_VERSION" in data:
                # like MOPAC2016.20.191M
                release, version = data["MOPAC_VERSION"].split(".", maxsplit=1)
                try:
                    t = datetime.datetime.strptime(version[0:-1], "%y.%j")
                    year = t.year
                    month = t.month
                    template = string.Template(self._bibliography["Stewart_2016"])
                    month = calendar.month_abbr[int(month)].lower()
                    citation = template.substitute(
                        month=month, version=version, year=year, release=release
                    )
                    self.references.cite(
                        raw=citation,
                        alias="mopac",
                        module="mopac_step",
                        level=1,
                        note="The principle MOPAC citation.",
                    )
                except Exception:
                    self.references.cite(
                        raw=self._bibliography["stewart_james_j_p_2022_6811510"],
                        alias="mopac",
                        module="mopac_step",
                        level=1,
                        note="The principle MOPAC citation.",
                    )
                break

        # Loop through our subnodes. Get the first real node
        node = self.subflowchart.get_node("1").next()
        first = 0
        n_node = 0
        while node:
            # Print the header for the node
            for value in node.description:
                printer.important(value)

            last = first + n_calculations[n_node]
            if last > len(out):
                logger.error("Could not find the MOPAC output for subjob {last + 1}/")
                node.analyze(data_sections=aux_data[first:last], out_sections=[])
            else:
                node.analyze(
                    data_sections=aux_data[first:last], out_sections=out[first:last]
                )
            first = last

            printer.normal("")

            node = node.next()
            n_node += 1

        if n_node > 1 and "CPU_TIME" in aux_data[-1]:
            text = f"MOPAC took a total of {t_total:.2f} s."
            printer.normal(str(__(text, **data, indent=self.indent)))

    def mopac_structure(self):
        """Create the input for the structure."""
        _, configuration = self.get_system_configuration(None)

        structure = ""
        atoms = configuration.atoms
        elements = atoms.symbols
        coordinates = atoms.get_coordinates(fractionals=False)
        if "freeze" in atoms:
            freeze = atoms["freeze"]
        else:
            freeze = [""] * len(elements)
        for element, xyz, frz in zip(elements, coordinates, freeze):
            x, y, z = xyz
            line = "{:2} {: 12.8f} {:d} {: 12.8f} {:d} {: 12.8f} {:d}\n".format(
                element,
                x,
                0 if "x" in frz else 1,
                y,
                0 if "y" in frz else 1,
                z,
                0 if "z" in frz else 1,
            )
            structure += line

        if configuration.periodicity == 3:
            # The three translation vectors
            element = "Tv"
            uvw = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
            XYZ = configuration.cell.to_cartesians(uvw)
            frz = 1 if self._lattice_opt else 0
            for xyz in XYZ:
                x, y, z = xyz
                line = (
                    f"{element:2} {x: 12.8f} {frz} {y: 12.8f} {frz} "
                    f"{z: 12.8f} {frz}\n"
                )
                structure += line

        return structure

    def parse_arc(self, filename="mopac.arc"):
        """Digest the ARC file and get the coordinates.

        Parameters
        ----------
        filename : str
            The name of the ARC file

        Returns
        -------
        coordinates : [n_atoms*[3]]
        """
        xyz = None
        cell_vectors = []
        with open(filename, "r") as fd:
            for line in fd:
                if "FINAL GEOMETRY OBTAINED" in line:
                    xyz = []
                    ii = 0
                    for line in fd:
                        if " &" not in line and " +" not in line and line[0] != "*":
                            ii += 1
                            if ii == 3:
                                break
                    for line in fd:
                        line = line.strip()
                        if line == "":
                            break
                        if line[0] != "*":
                            symbol, x, fx, y, fy, z, fz = line.split()
                            if symbol == "Tv":
                                cell_vectors.append(
                                    [float(x), float(y), float(z)]
                                )  # yapf: disable
                            else:
                                xyz.append([float(x), float(y), float(z)])
        return xyz, cell_vectors

    def parse_aux(self, lines):
        """Digest a section of the aux file"""

        properties = mopac_step.metadata["results"]
        trans = str.maketrans("Dd", "Ee")

        data = {}
        lineno = -1
        nlines = len(lines)

        spin_polarized = False

        while True:
            lineno += 1
            if lineno >= nlines:
                break
            line = lines[lineno].strip()
            if line[0] == "#":
                continue
            if "=" not in line:
                raise RuntimeError("Problem parsing MOPAC aux file: '" + line + "'")
            key, rest = line.split("=", maxsplit=1)
            if key[-1] == "]":
                name, size = key[0:-1].split("[")
                size = int(size.lstrip("0"))
                if ":" in name:
                    name, units = name.split(":")

                if name not in properties:
                    logger.warning("Property '{}' not recognized.".format(name))
                    kind = "string"
                else:
                    kind = properties[name]["type"]
                    if "units" in properties[name]:
                        data[name + ",units"] = properties[name]["units"]

                if name == "NUM_ALPHA_ELECTRONS":
                    spin_polarized = True

                # Bug workaround
                # Sometimes MOPAC does not write out the MO occupancies
                if name == "MOLECULAR_ORBITAL_OCCUPANCIES":
                    tmp_line = lines[lineno + 1].strip()
                    if tmp_line[0].isalpha():
                        continue
                # end of workaround

                # Check for floating point numbers run together
                if kind == "float":
                    values = []
                    for value in rest.split():
                        tmp = value.split(".")
                        if len(tmp) <= 2:
                            values.append(value)
                        else:
                            # Run together ... lets see how many decimals
                            n_decimals = len(tmp[-1])
                            # and before the decimal
                            n_digits = len(tmp[-2]) - n_decimals
                            n = n_digits + 1 + n_decimals
                            n_values = len(tmp) - 1
                            # blanks at front have been stripped, so count back
                            start = 0
                            end = len(value) - (n_values - 1) * n
                            while start < len(value):
                                values.append(value[start:end])
                                start = end
                                end += n
                    tmp = values
                else:
                    tmp = rest.split()

                # AUX file contains alpha and beta orbitals, that's why we are
                # doubling it
                if name == "MICROSTATE_CONFIGURATIONS" and spin_polarized:
                    size = size * 2

                while len(tmp) < size:
                    lineno += 1
                    line = lines[lineno].strip()
                    if line[0] != "#":
                        tmp.extend(line.split())

                if kind == "integer":
                    values = []
                    for value in tmp:
                        values.append(int(value))
                elif kind == "float":
                    values = []
                    for value in tmp:
                        values.append(float(value.translate(trans)))
                else:
                    values = tmp

                if "UPDATED" in name:
                    if name not in data:
                        data[name] = []
                    data[name].append(values)
                else:
                    if (
                        name in properties
                        and properties[name]["dimensionality"] == "scalar"
                    ):
                        if not (units == "ARBITRARY_UNITS" and name in data):
                            data[name] = values[0]
                    else:
                        data[name] = values
            else:
                if ":" in key:
                    name, units = key.split(":")
                else:
                    name = key

                if name not in properties:
                    logger.warning("Property '{}' not recognized.".format(name))
                    kind = "string"
                else:
                    kind = properties[name]["type"]
                if "units" in properties[name]:
                    data[name + ",units"] = properties[name]["units"]
                if kind == "integer":
                    value = int(rest)
                elif kind == "float":
                    value = float(self._sanitize_value(rest.strip()))
                else:
                    value = rest.strip('"')

                if "UPDATED" in name:
                    if name not in data:
                        data[name] = []
                    data[name].append(value)
                else:
                    data[name] = value

        return data

    def _sanitize_value(self, value):
        regex = r"^([-+]?[.0-9]+)([EeDd]*)([-+][0-9]+)$"
        subs = r"\1E\3"
        ret = float(re.sub(regex, subs, value))
        return ret



# -*- coding: utf-8 -*-

"""The graphical part of a MOPAC step in SEAMM"""

import logging
import seamm

logger = logging.getLogger(__name__)


class TkMOPAC(seamm.TkNode):
    """The node_class is the class of the 'real' node that this
    class is the Tk graphics partner for
    """

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        namespace="org.molssi.seamm.mopac.tk",
        x=120,
        y=20,
        w=200,
        h=50,
        my_logger=logger,
    ):
        """Initialize the graphical Tk node for MOPAC

        Keyword arguments:
        """
        self.namespace = namespace

        # Set the logging level for this module if requested
        # if 'tk_mopac_log_level' in self.options:
        #     logger.setLevel(self.options.tk_mopac_log_level)
        #     logger.critical(
        #         'Set log level to {}'.format(self.options.tk_mopac_log_level)
        #     )

        # Call the constructor for the energy
        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
        )

        self.create_dialog()

    def create_dialog(self):
        """Create the dialog!"""
        frame = super().create_dialog("Edit MOPAC Step")

        # make it large!
        sw = self.dialog.winfo_screenwidth()
        sh = self.dialog.winfo_screenheight()
        w = int(0.9 * sw)
        h = int(0.8 * sh)
        x = int(0.05 * sw / 2)
        y = int(0.1 * sh / 2)

        self.dialog.geometry("{}x{}+{}+{}".format(w, h, x, y))

        self.tk_subflowchart = seamm.TkFlowchart(
            master=frame, namespace=self.namespace, flowchart=self.node.subflowchart
        )
        self.tk_subflowchart.draw()

    def right_click(self, event):
        """Probably need to add our dialog..."""

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)



# -*- coding: utf-8 -*-

"""The graphical part of a MOPAC Energy node"""

import mopac_step
import seamm_widgets as sw
import tkinter as tk
import tkinter.ttk as ttk


class TkOptimization(mopac_step.TkEnergy):
    def __init__(
        self, tk_flowchart=None, node=None, canvas=None, x=120, y=20, w=200, h=50
    ):
        """Initialize a node

        Keyword arguments:
        """
        super().__init__(
            tk_flowchart=tk_flowchart, node=node, canvas=canvas, x=x, y=y, w=w, h=h
        )
        self.mopac_parameters = self.node.parent.parameters

    def right_click(self, event):
        """Probably need to add our dialog..."""

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def create_dialog(self, title="MOPAC Optimization"):
        """Create the dialog!"""
        self.logger.debug("Creating the dialog")

        frame = super().create_dialog(title=title)

        P = self.node.parameters

        # Create the optimization widgets
        oframe = self["optimization frame"] = ttk.LabelFrame(
            frame, text="Optimization", labelanchor=tk.N
        )
        row = 0
        for key in mopac_step.OptimizationParameters.parameters:
            self[key] = P[key].widget(oframe)
            row += 1
        for key in mopac_step.structure_handling_parameters:
            self[key] = P[key].widget(oframe)
            row += 1

        # Set the callbacks for changes
        for widget in ("method", "convergence", "LatticeOpt"):
            w = self[widget]
            w.combobox.bind(
                "<<ComboboxSelected>>", self.reset_optimization_frame, add="+"
            )
            w.combobox.bind("<Return>", self.reset_optimization_frame, add="+")
            w.combobox.bind("<FocusOut>", self.reset_optimization_frame, add="+")

        self.logger.debug("Finished creating the dialog")

        return frame

    def reset_dialog(self, widget=None):
        row = super().reset_dialog()

        self["optimization frame"].grid(row=row, column=0, sticky=tk.EW)
        row += 1

        # And the widgets in our frame
        self.reset_optimization_frame()

        return row

    def reset_optimization_frame(self, widget=None):
        """Layout the widgets in the optimization frame

        We'll let 'TkEnergy' layout the initial set of widgets,
        then add the extra widgets for controlling optimization
        """
        frame = self["optimization frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        convergence = self["convergence"].get()
        method = self["method"].get()
        lattice_opt = self["LatticeOpt"].get()

        widgets = []
        widgets_2 = []
        row = 0

        self["LatticeOpt"].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
        widgets.append(self["LatticeOpt"])
        row += 1

        if lattice_opt == "Yes":
            self["pressure"].grid(row=row, column=1, sticky=tk.EW)
            row += 1

        self["method"].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
        widgets.append(self["method"])
        row += 1

        self["cycles"].grid(row=row, column=1, sticky=tk.EW)
        widgets_2.append(self["cycles"])
        row += 1

        if convergence not in ("normal", "precise"):
            self["gnorm"].grid(row=row, column=1, sticky=tk.EW)
            widgets_2.append(self["gnorm"])
            row += 1
        if method[0:2] == "EF" or method[0] == "$":
            self["recalc"].grid(row=row, column=1, sticky=tk.EW)
            widgets_2.append(self["recalc"])
            row += 1
            self["dmax"].grid(row=row, column=1, sticky=tk.EW)
            widgets_2.append(self["dmax"])
            row += 1

        for key in ("structure handling", "configuration name"):
            self[key].grid(row=row, column=0, columnspan=2, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        sw.align_labels(widgets, sticky=tk.E)
        sw.align_labels(widgets_2, sticky=tk.E)

        frame.columnconfigure(0, minsize=100)

        return row



# -*- coding: utf-8 -*-

"""Top-level package for MOPAC step."""

# Bring up the classes so that they appear to be directly in
# the package.

from .mopac_step import MOPACStep  # noqa: F401
from .mopac import MOPAC  # noqa: F401
from .tk_mopac import TkMOPAC  # noqa: F401
from .energy_step import EnergyStep  # noqa: F401
from .energy import Energy  # noqa: F401
from .energy_parameters import EnergyParameters  # noqa: F401
from .tk_energy import TkEnergy  # noqa: F401
from .forceconstants_step import ForceconstantsStep  # noqa: F401
from .forceconstants import Forceconstants  # noqa: F401
from .forceconstants_parameters import ForceconstantsParameters  # noqa: F401
from .tk_forceconstants import TkForceconstants  # noqa: F401
from .optimization_step import OptimizationStep  # noqa: F401
from .optimization import Optimization  # noqa: F401
from .optimization_parameters import OptimizationParameters  # noqa: F401
from .tk_optimization import TkOptimization  # noqa: F401
from .ir_step import IRStep  # noqa: F401
from .ir import IR  # noqa: F401
from .ir_parameters import IRParameters  # noqa: F401
from .tk_ir import TkIR  # noqa: F401
from .thermodynamics_step import ThermodynamicsStep  # noqa: F401
from .thermodynamics import Thermodynamics  # noqa: F401
from .thermodynamics_parameters import (  # noqa: F401
    ThermodynamicsParameters,
)
from .tk_thermodynamics import TkThermodynamics  # noqa: F401

from .metadata import metadata  # noqa: F401

# Handle versioneer
from ._version import get_versions

# Parameters used for handling the structure if it is changed.
structure_handling_parameters = {
    "structure handling": {
        "default": "be put in a new configuration",
        "kind": "enum",
        "default_units": "",
        "enumeration": (
            "overwrite the current configuration",
            "be put in a new configuration",
        ),
        "format_string": "s",
        "description": "Optimized structure will",
        "help_text": (
            "Whether to overwrite the current configuration, or create a new "
            "configuration or system and configuration for the new structure"
        ),
    },
    "configuration name": {
        "default": "optimized with <Hamiltonian>",
        "kind": "string",
        "default_units": "",
        "enumeration": (
            "optimized with <Hamiltonian>",
            "keep current name",
            "use SMILES string",
            "use Canonical SMILES string",
            "use configuration number",
        ),
        "format_string": "s",
        "description": "Configuration name:",
        "help_text": "The name for the new configuration",
    },
}

__author__ = """Paul Saxe"""
__email__ = "psaxe@molssi.org"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



# -*- coding: utf-8 -*-
"""Control parameters for a MOPAC optimization calculation
"""

import logging
import mopac_step

logger = logging.getLogger(__name__)


class OptimizationParameters(mopac_step.EnergyParameters):
    """The control parameters for optimization in MOPAC"""

    parameters = {
        "method": {
            "default": "default",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "default",
                "EF -- eigenvector following",
                "BFGS -- Broyden-Fletcher-Goldfarb-Shanno algorithm",
                "L-BFGS -- smaller memory BFGS for larger systems",
                "TS -- transition state with EF method",
                "SIGMA -- transition state with McIver-Komornicki method",
                "NLLSQ -- nonlinear least squares of gradient, Bartel's method",
            ),
            "format_string": "s",
            "description": "Optimization algorithm:",
            "help_text": ("The optimization algorithm to use."),
        },
        "gnorm": {
            "default": "1.0",
            "kind": "float",
            "default_units": "kcal/mol/Å",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Convergence criterion:",
            "help_text": (
                "The convergence criterion for the optimization. "
                "This is controlled in part by the overall "
                "convergence criterion:"
                "* 'normal' is the default level, 1.0 "
                " kcal/mol/Å."
                "* 'precise' tightens up the criterion by a factor "
                "of 100."
                "* 'absolute' uses the value set here."
            ),
        },
        "cycles": {
            "default": "unlimited",
            "kind": "integer",
            "default_units": "",
            "enumeration": ("unlimited",),
            "format_string": "",
            "description": "Maximum steps:",
            "help_text": (
                "The maximum number of steps to take "
                "in the optimization. 'unlimited' means "
                "there is no definite limit; however, the "
                "time-limit of the job will stop the job."
            ),
        },
        "recalc": {
            "default": "never",
            "kind": "integer",
            "default_units": "",
            "enumeration": ("never",),
            "format_string": "",
            "description": "Recalculate Hessian:",
            "help_text": (
                "How often to recalculate the Hessian "
                "(in steps) "
                "when using the EF method. Smaller values "
                "help convergence but are expensive. 10 "
                "with a trust radius of 0.1 is useful for "
                "difficult cases, while 1 and a trust radius "
                "of 0.05 will always find a stationary point "
                "if one exists."
            ),
        },
        "dmax": {
            "default": "0.2",
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Trust radius:",
            "help_text": (
                "The trust radius in the EF method. The "
                "algorithm updates this based on progress. "
                "The default is normally fine, but in some "
                "difficult cases reducing to 0.1 or 0.05 "
                "may be useful."
            ),
        },
        "LatticeOpt": {
            "default": "Yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("Yes", "No"),
            "format_string": "",
            "description": "Optimize the cell (if periodic):",
            "help_text": "Allow the lattice vectors to change during optimization.",
        },
        "pressure": {
            "default": 0.0,
            "kind": "float",
            "default_units": "GPa",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Pressure:",
            "help_text": ("The applied pressure."),
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={
                **OptimizationParameters.parameters,
                **mopac_step.structure_handling_parameters,
                **defaults,
            },
            data=data,
        )



# -*- coding: utf-8 -*-

"""Control parameters for a MOPAC vibrational (Forceconstants/Raman) calculation
"""

import logging
import mopac_step

logger = logging.getLogger(__name__)


class ForceconstantsParameters(mopac_step.EnergyParameters):
    """The control parameters for creating a structure from SMILES"""

    parameters = {
        "what": {
            "default": "full Hessian",
            "kind": "enum",
            "default_units": "",
            "enumeration": ("full Hessian", "atom part only", "cell part only"),
            "format_string": "",
            "group": "",
            "description": "Output:",
            "help_text": "What part(s) of the Hessian to output",
        },
        "stepsize": {
            "default": "0.01",
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": ".3f",
            "description": "Strain step size:",
            "help_text": (
                "The step size to use straining the system in the finite difference "
                "step."
            ),
        },
        "two-sided_atoms": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "group": "",
            "description": "Two-sided atom displacements:",
            "help_text": (
                "Whether to use two-sided finite differences for atom portion."
            ),
        },
        "two-sided_cell": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "group": "",
            "description": "Two-sided strain differences:",
            "help_text": (
                "Whether to use two-sided finite differences for cell portion."
            ),
        },
        "atom_units": {
            "default": "N/m",
            "kind": "str",
            "default_units": "",
            "enumeration": ("N/m", "kJ/mol/Å^2", "kcal/mol/Å^2", "eV/Å^2"),
            "format_string": "",
            "group": "",
            "description": "Units for atom part:",
            "help_text": (
                "What units to use for the atom-atom and atom-cell parts of the Hessian"
            ),
        },
        "cell_units": {
            "default": "GPa",
            "kind": "str",
            "default_units": "",
            "enumeration": ("GPa", "atm"),
            "format_string": "",
            "group": "",
            "description": "Units for cell part:",
            "help_text": "What units to use for the cell-cell part of the Hessian",
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={
                **ForceconstantsParameters.parameters,
                **mopac_step.structure_handling_parameters,
                **defaults,
            },
            data=data,
        )



# -*- coding: utf-8 -*-

"""The graphical part of a MOPAC vibrational node"""

import mopac_step
import seamm_widgets as sw
import tkinter as tk
import tkinter.ttk as ttk


class TkIR(mopac_step.TkEnergy):
    def __init__(
        self, tk_flowchart=None, node=None, canvas=None, x=120, y=20, w=200, h=50
    ):
        """Initialize a node

        Keyword arguments:
        """
        super().__init__(
            tk_flowchart=tk_flowchart, node=node, canvas=canvas, x=x, y=y, w=w, h=h
        )

    def create_dialog(self, title="MOPAC Infrared (Vibrational) Spectrum"):
        """Create the dialog!"""
        self.logger.debug("Creating the dialog")
        frame = super().create_dialog(title=title)

        P = self.node.parameters

        # Create the vibrational widgets
        vframe = self["vibrational frame"] = ttk.LabelFrame(
            frame, text="Vibrational Analysis", labelanchor=tk.N
        )
        row = 0
        widgets = []
        for key in mopac_step.IRParameters.parameters:
            self[key] = P[key].widget(vframe)
            self[key].grid(row=row, column=0, sticky=tk.EW)
            widgets.append(self[key])
            row += 1
        sw.align_labels(widgets, sticky=tk.E)

        # Create the structure-handling widgets
        sframe = self["structure frame"] = ttk.LabelFrame(
            frame, text="Configuration Handling", labelanchor=tk.N
        )
        row = 0
        widgets = []
        for key in ("structure handling", "configuration name"):
            self[key] = P[key].widget(sframe)
            self[key].grid(row=row, column=0, sticky=tk.EW)
            widgets.append(self[key])
            row += 1
        sw.align_labels(widgets)

        self.logger.debug("Finished creating the dialog")

    def reset_dialog(self, widget=None):
        """Layout the widgets in the main frame

        We'll let 'TkEnergy' layout the initial set of widgets,
        then add the extra widgets for controlling optimization
        """
        row = super().reset_dialog()

        # And our frame
        self["vibrational frame"].grid(row=row, column=0, sticky=tk.EW, pady=10)
        row += 1

        # And how to handle the configuration
        self["structure frame"].grid(row=row, column=0, sticky=tk.EW, pady=10)
        row += 1

        return row



# -*- coding: utf-8 -*-

"""Setup and run MOPAC"""

import logging
from pathlib import Path
import textwrap
import traceback

import mopac_step
import seamm
import seamm_util.printing as printing
from seamm_util import units_class
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("mopac")


class Optimization(mopac_step.Energy):
    def __init__(self, flowchart=None, title="Optimization", extension=None):
        """Initialize the node"""

        logger.debug("Creating Optimization {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self._calculation = "optimization"
        self._model = None
        self._metadata = mopac_step.metadata
        self.parameters = mopac_step.OptimizationParameters()
        self.description = "A structural optimization"

    def description_text(self, P=None):
        """Prepare information about what this node will do"""

        if not P:
            P = self.parameters.values_to_dict()

        # The energy part of the description
        tmp = super().description_text(P)
        energy_description = textwrap.dedent("\n".join(tmp.splitlines()[1:]))

        # Hamiltonian followed by convergence
        text = "Geometry optimization with the "
        if P["method"] == "default":
            text += "default optimizer (EF for small systems," " L-BFGS for larger)."
        elif P["method"][0:1] == "EF":
            text += "eigenvector following (EF) method."
        elif P["method"][0:3] == "BFGS":
            text += "BFGS method."
        elif P["method"][0:5] == "L-BFGS":
            text += "L-BFGS small memory version of the BFGS method."
        else:
            text += "optimization method determined at runtime by '{method}'."

        if P["LatticeOpt"]:
            pressure = P["pressure"]
            if isinstance(pressure, units_class):
                pressure = f"{pressure:~P}"
            text += (
                " The cell for periodic systems will also be optimized, with an "
                f"applied pressure of {pressure}."
            )
        else:
            text += " The cell for periodic systems will not be optimized."

        if P["gnorm"] == "default":
            text += " The geometrical convergence is the default of " "1.0 kcal/mol/Å."
        elif P["gnorm"][0] == "$":
            text += (
                " The geometrical convergence will be determined "
                "at runtime by '{gnorm}'."
            )
        else:
            text += " The geometrical convergence is {gnorm} kcal/mol/Å."

        # Put in the description of the energy calculation
        text += "\n\nThe energy and forces will be c" + energy_description[1:]
        text += "\n\n"

        text += "The optimized structures will {structure handling} "

        confname = P["configuration name"]
        if confname == "use SMILES string":
            text += "using SMILES as its name."
        elif confname == "use Canonical SMILES string":
            text += "using canonical SMILES as its name."
        elif confname == "keep current name":
            text += "keeping the current name."
        elif confname == "optimized with <Hamiltonian>":
            text += "with 'optimized with {hamiltonian}' as its name."
        elif confname == "use configuration number":
            text += "using the index of the configuration (1, 2, ...) as its name."
        else:
            confname = confname.replace("<Hamiltonian>", P["hamiltonian"])
            text += f"with '{confname}' as its name."

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self):
        """Get the input for an optimization MOPAC"""

        references = self.parent.references
        bibliography = self.parent._bibliography

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        system, configuration = self.get_system_configuration(None)

        # Let parent know about cell optimization
        self.parent._lattice_opt = P["LatticeOpt"]

        # Get the inputs from the energy class. This also sets the description properly.
        inputs = super().get_input()

        # Remove the 1SCF keyword from the energy setup
        # 'keywords' is a reference, so will change in situ.
        keywords, _, _ = inputs[0]
        if "1SCF" in keywords:
            keywords.remove("1SCF")

        # Pressure
        if configuration.periodicity == 3 and P["LatticeOpt"]:
            pressure = P["pressure"]
            pressure = pressure.to("GPa").magnitude
            keywords.append(f"P={pressure}GPa")

        # and the optimization-specific parts
        method = P["method"]
        if method == "default":
            pass
        elif method[0:2] == "EF":
            keywords.append("EF")
            if P["recalc"] != "never":
                keywords.append(P["recalc"])
            if str(P["dmax"]) != self.parameters["dmax"].default:
                keywords.append("DMAX={}".format(P["dmax"]))
            references.cite(
                raw=bibliography["Baker_1986"],
                alias="Baker_1986",
                module="mopac_step",
                level=1,
                note="Eigenvector-following minimizer.",
            )
        elif method[0:4] == "BFGS":
            keywords.append("BFGS")
            references.cite(
                raw=bibliography["Broyden_1970"],
                alias="Broyden_1970",
                module="mopac_step",
                level=1,
                note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
            )
            references.cite(
                raw=bibliography["Fletcher_1970"],
                alias="Fletcher_1970",
                module="mopac_step",
                level=1,
                note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
            )
            references.cite(
                raw=bibliography["Goldfarb_1970"],
                alias="Goldfarb_1970",
                module="mopac_step",
                level=1,
                note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
            )
            references.cite(
                raw=bibliography["Shanno_1970"],
                alias="Shanno_1970",
                module="mopac_step",
                level=1,
                note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
            )
            references.cite(
                raw=bibliography["Thiel_1988"],
                alias="Thiel_1988",
                module="mopac_step",
                level=1,
                note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
            )
        elif method[0:6] == "L-BFGS":
            keywords.append("LBFGS")
            references.cite(
                raw=bibliography["Nocedal_1980"],
                alias="Nocedal_1980",
                module="mopac_step",
                level=1,
                note="Limited-memory BFGS (L-BFGS) minimizer.",
            )
        elif method[0:2] == "TS":
            keywords.append("TS")
            references.cite(
                raw=bibliography["Baker_1986"],
                alias="Baker_1986",
                module="mopac_step",
                level=1,
                note="Eigenvector-following minimizer for transition states.",
            )
        elif method[0:5] == "SIGMA":
            keywords.append("SIGMA")
            references.cite(
                raw=bibliography["McIver_1971"],
                alias="McIver_1971",
                module="mopac_step",
                level=1,
                note="Gradient-norm minimizer for transition states.",
            )
            references.cite(
                raw=bibliography["McIver_1972"],
                alias="McIver_1972",
                module="mopac_step",
                level=1,
                note="Gradient-norm minimizer for transition states.",
            )
        elif method[0:5] == "NLLSQ":
            keywords.append("NLLSQ")
            references.cite(
                raw=bibliography["Bartels_1972"],
                alias="Bartels_1972",
                module="mopac_step",
                level=1,
                note="NLLSQ gradient-norm minimizer for transition states.",
            )
        else:
            text = "Don't recognize optimization method '{}'".format(P["method"])
            logger.critical(text)
            raise RuntimeError(text)

        if P["cycles"] != "unlimited":
            keywords.append("CYCLES={}".format(P["cycles"]))
        if P["convergence"] not in ("normal", "precise"):
            if P["gnorm"] != self.parameters["gnorm"].default:
                keywords.append("GNORM={}".format(P["gnorm"]))

        return inputs

    def analyze(self, indent="", data_sections=[], out_sections=[], table=None):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """

        # Get the parameters used
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        system, starting_configuration = self.get_system_configuration(None)

        # Get the data.
        data = data_sections[0]

        # If the optimizer used was the default, put in the correct citations

        references = self.parent.references
        bibliography = self.parent._bibliography

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        if P["method"] == "default":
            tmp = "\n".join(out_sections[0])
            if (
                "GEOMETRY OPTIMISED USING EIGENVECTOR FOLLOWING (EF)" in tmp
                or "Geometry optimization using EF" in tmp
            ):
                opt_method = "EF -- eigenvector following"
                references.cite(
                    raw=bibliography["Baker_1986"],
                    alias="Baker_1986",
                    module="mopac_step",
                    level=1,
                    note="Eigenvector-following minimizer.",
                )
            elif (
                "Geometry optimization using BFGS" in tmp or "SATISFIED IN BFGS" in tmp
            ):
                opt_method = "BFGS -- Broyden-Fletcher-Goldfarb-Shanno algorithm"
                references.cite(
                    raw=bibliography["Broyden_1970"],
                    alias="Broyden_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Fletcher_1970"],
                    alias="Fletcher_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Goldfarb_1970"],
                    alias="Goldfarb_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Shanno_1970"],
                    alias="Shanno_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Thiel_1988"],
                    alias="Thiel_1988",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
            elif (
                "Geometry optimization using L-BFGS" in tmp
                or "SATISFIED IN L-BFGS" in tmp
            ):
                opt_method = (
                    "L-BFGS -- Limited memory Broyden-Fletcher-Goldfarb-Shanno "
                    "algorithm"
                )
                references.cite(
                    raw=bibliography["Broyden_1970"],
                    alias="Broyden_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Fletcher_1970"],
                    alias="Fletcher_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Goldfarb_1970"],
                    alias="Goldfarb_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Shanno_1970"],
                    alias="Shanno_1970",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
                references.cite(
                    raw=bibliography["Thiel_1988"],
                    alias="Thiel_1988",
                    module="mopac_step",
                    level=1,
                    note="Broyden-Fletcher-Goldfarb-Shanno (BFGS) minimizer.",
                )
            else:
                logger.warning("Could not find which minimizer was used!")
                opt_method = "default -- but could not find which was used"
        else:
            opt_method = P["method"]

        # The results
        if "NUMBER_SCF_CYCLES" in data_sections[0]:
            text = (
                f"The geometry optimization using {opt_method} -- converged in "
                "{NUMBER_SCF_CYCLES} iterations."
            )
        else:
            data["NUMBER_SCF_CYCLES"] = len(data["HEAT_OF_FORM_UPDATED"])
            data["HEAT_OF_FORMATION"] = data["HEAT_OF_FORM_UPDATED"][-1]
            data["GRADIENT_NORM"] = data["GRADIENT_UPDATED"][-1]
            text = (
                f"The geometry optimization using {opt_method} -- did not converge in "
                "{NUMBER_SCF_CYCLES} steps. The following results are for the final "
                "structure.\n"
            )

        # Update the structure
        if "ATOM_X_OPT" in data or "ATOM_X_UPDATED" in data:

            periodicity = starting_configuration.periodicity
            if (
                "structure handling" in P
                and P["structure handling"] == "be put in a new configuration"
            ):
                configuration = system.create_configuration(
                    periodicity=periodicity,
                    atomset=starting_configuration.atomset,
                    bondset=starting_configuration.bondset,
                    cell_id=starting_configuration.cell_id,
                )
                configuration.charge = starting_configuration.charge
                configuration.spin_multiplicity = (
                    starting_configuration.spin_multiplicity
                )
            else:
                configuration = starting_configuration

            if periodicity != 0 and P["LatticeOpt"]:
                if "TRANS_VECTS" in data:
                    vectors = data["TRANS_VECTS"]
                    lattice = [
                        [vectors[0], vectors[1], vectors[2]],
                        [vectors[3], vectors[4], vectors[5]],
                        [vectors[6], vectors[7], vectors[8]],
                    ]
                    configuration.cell.from_vectors(lattice)
                elif "TRANS_VECTS_UPDATED" in data:
                    vectors = data["TRANS_VECTS_UPDATED"][-1]
                    lattice = [
                        [vectors[0], vectors[1], vectors[2]],
                        [vectors[3], vectors[4], vectors[5]],
                        [vectors[6], vectors[7], vectors[8]],
                    ]
                    configuration.cell.from_vectors(lattice)
                else:
                    logger.warning(
                        "Expected updated lattice vectors, but did not find!"
                    )
            xyz = []
            if "ATOM_X_OPT" in data:
                it = iter(data["ATOM_X_OPT"])
            else:
                it = iter(data["ATOM_X_UPDATED"][-1])
            for x in it:
                xyz.append([float(x), float(next(it)), float(next(it))])
            configuration.atoms.set_coordinates(xyz, fractionals=False)

            # And the name of the configuration.
            if "configuration name" in P:
                if P["configuration name"] == "use SMILES string":
                    configuration.name = configuration.smiles
                elif P["configuration name"] == "use Canonical SMILES string":
                    configuration.name = configuration.canonical_smiles
                elif P["configuration name"] == "keep current name":
                    pass
                elif P["configuration name"] == "optimized with <Hamiltonian>":
                    configuration.name = f"optimized with {P['hamiltonian']}"
                else:
                    confname = P["configuration name"]
                    confname = confname.replace("<Hamiltonian>", P["hamiltonian"])
                    configuration.name = confname

        # Write the structure out for viewing.
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        #  MMCIF file has bonds
        try:
            path = directory / "optimized.mmcif"
            path.write_text(configuration.to_mmcif_text())
        except Exception:
            message = "Error creating the mmcif file\n\n" + traceback.format_exc()
            logger.warning(message)
        # CIF file has cell
        if configuration.periodicity == 3:
            try:
                path = directory / "optimized.cif"
                path.write_text(configuration.to_cif_text())
            except Exception:
                message = "Error creating the cif file\n\n" + traceback.format_exc()
                logger.warning(message)

        printer.normal(__(text, **data, indent=self.indent + 4 * " "))

        super().analyze(
            indent=indent,
            data_sections=data_sections,
            out_sections=out_sections,
            table=table,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""mopac_step
A SEAMM plug-in to setup, run and analyze semiempirical calculations with MOPAC
"""

import sys
from setuptools import setup, find_packages
import versioneer

short_description = __doc__.splitlines()[1]

# from https://github.com/pytest-dev/pytest-runner#conditional-requirement
needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as fd:
    requirements = fd.read()

setup(
    # Descriptive entries which should always be present
    name='mopac_step',
    author="Paul Saxe",
    author_email='psaxe@molssi.org',
    description=short_description,
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license='BSD-3-Clause',
    url='https://github.com/molssi-seam/mopac_step',
    packages=find_packages(include=['mopac_step']),
    include_package_data=True,

    # Allows `setup.py test` to work correctly with pytest
    setup_requires=[] + pytest_runner,

    # Required packages, pulls from pip if needed; do not use for Conda
    # deployment
    install_requires=requirements,

    test_suite='tests',

    # Valid platforms your code works on, adjust to your flavor
    platforms=['Linux',
               'Mac OS-X',
               'Unix',
               'Windows'],

    # Manual control if final package is compressible or not, set False to
    # prevent the .egg from being made
    zip_safe=True,

    keywords=['SEAMM', 'plug-in', 'flowchart', 'forcefield', 'MOPAC',
              'semiempirical', 'orbitals'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    entry_points={
        'console_scripts': [
            'mopac-step-installer=mopac_step.__main__:run',
        ],
        'org.molssi.seamm': [
            'MOPAC = mopac_step:MOPACStep',
        ],
        'org.molssi.seamm.tk': [
            'MOPAC = mopac_step:MOPACStep',
        ],
        'org.molssi.seamm.mopac': [
            'Energy = mopac_step:EnergyStep',
            'Optimization = mopac_step:OptimizationStep',
            'IR Spectrum = mopac_step:IRStep',
            'Thermodynamics = mopac_step:ThermodynamicsStep',
            'Forceconstants = mopac_step:ForceconstantsStep',
        ],
        'org.molssi.seamm.mopac.tk': [
            'Energy = mopac_step:EnergyStep',
            'Optimization = mopac_step:OptimizationStep',
            'IR Spectrum = mopac_step:IRStep',
            'Thermodynamics = mopac_step:ThermodynamicsStep',
            'Forceconstants = mopac_step:ForceconstantsStep',
        ],
    },
)



# -*- coding: utf-8 -*-

"""Installer for the MOPAC plug-in.

This handles any further installation needed after installing the Python
package `mopac-step`.
"""

import logging
from pathlib import Path
import pkg_resources
import subprocess

import seamm_installer

logger = logging.getLogger(__name__)


class Installer(seamm_installer.InstallerBase):
    """Handle further installation needed after installing mopac-step.

    The Python package `MOPAC-step` should already be installed, using `pip`,
    `conda`, or similar. This plug-in-specific installer then checks for the
    MOPAC executable, installing it if needed, and registers its
    location in seamm.ini.

    There are a number of ways to determine which are the correct MOPAC
    executables to use. The aim of this installer is to help the user locate
    the executables. There are a number of possibilities:

    1. The correct executables are already available.

        1. If they are already registered in `seamm.ini` there is nothing else
           to do.
        2. They may be in the current path, in which case they need to be added
           to `seamm.ini`.
        3. If a module system is in use, a module may need to be loaded to give
           access to MOPAC.
        3. They cannot be found automatically, so the user needs to locate the
           executables for the installer.

    2. MOPAC is not installed on the machine. In this case they can be
       installed in a Conda environment. There is one choice:

        1. They can be installed in a separate environment, `seamm-mopac` by
           default.
    """

    def __init__(self, logger=logger):
        # Call the base class initialization, which sets up the commandline
        # parser, amongst other things.
        super().__init__(logger=logger)

        logger.debug("Initializing the MOPAC installer object.")

        self.section = "mopac-step"
        self.path_name = "mopac-path"
        self.executables = ["mopac"]
        self.resource_path = Path(pkg_resources.resource_filename(__name__, "data/"))
        # What Conda environment is the default?
        data = self.configuration.get_values(self.section)
        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-mopac"

        # The environment.yaml file for Conda installations.
        path = Path(pkg_resources.resource_filename(__name__, "data/"))
        logger.debug(f"data directory: {path}")
        self.environment_file = path / "seamm-mopac.yml"

    def exe_version(self, path):
        """Get the version of the MOPAC executable.

        Parameters
        ----------
        path : pathlib.Path
            Path to the executable.

        Returns
        -------
        str
            The version reported by the executable, or 'unknown'.
        """
        try:
            result = subprocess.run(
                [str(path), "--version"],
                stdin=subprocess.DEVNULL,
                capture_output=True,
                text=True,
            )
        except Exception:
            version = "unknown"
        else:
            version = "unknown"
            lines = result.stdout.splitlines()
            for line in lines:
                line = line.strip()
                tmp = line.split()
                if len(tmp) >= 2:
                    version = tmp[2]
                    break

        return version
